@@ -1,6 +1,4 @@
 
-local floor = math.floor
-
 require('lualibs')
 
 ------------------------------------------------------------------------
@@ -8,32 +6,56 @@ require('lualibs')
 --    ext_... : called from \directlua{}
 --    int_... : called from other Lua codes, but not from \directlua{}
 --    (other)     : only called from this file
-function luatexja.error(s,t)
-   tex.error('LuaTeX-ja error: ' .. s ,t) 
-end
 function luatexja.load_module(name)
-   if not package.loaded['luatexja.' .. name] then
-      local fn = 'ltj-' .. name .. '.lua'
-      local found = kpse.find_file(fn, 'tex')
-      if not found then
-	 luatexja.error("File `" .. fn .. "' not found", 
-			{'This file ' .. fn .. ' is required for LuaTeX-ja.', 'Please check your installation.'})
-      else 
-	 texio.write('(' .. found .. ')\n')
-	 dofile(found)
-      end
-   end
+   require('ltj-' .. name.. '.lua')
 end
 function luatexja.load_lua(fn)
    local found = kpse.find_file(fn, 'tex')
    if not found then
-      error("File `" .. fn .. "' not found")
+      tex.error("LuaTeX-ja error: File `" .. fn .. "' not found")
    else 
-      texio.write('(' .. found .. ')\n')
+      texio.write_nl('(' .. found .. ')')
       dofile(found)
    end
 end
 
+--- 以下は全ファイルで共有される定数
+local icflag_table = {}
+luatexja.icflag_table = icflag_table
+icflag_table.ITALIC       = 1
+icflag_table.PACKED       = 2
+icflag_table.KINSOKU      = 3
+icflag_table.FROM_JFM     = 6
+-- FROM_JFM: 4, 5, 6, 7, 8 →優先度高
+-- 6 が標準
+icflag_table.KANJI_SKIP   = 9
+icflag_table.XKANJI_SKIP  = 10
+icflag_table.PROCESSED    = 11
+icflag_table.IC_PROCESSED = 12
+icflag_table.BOXBDD       = 15
+icflag_table.PROCESSED_BEGIN_FLAG = 32
+
+local stack_table_index = {}
+luatexja.stack_table_index = stack_table_index
+stack_table_index.PRE  = 0x200000 -- characterごと
+stack_table_index.POST = 0x400000 -- characterごと
+stack_table_index.KCAT = 0x600000 -- characterごと
+stack_table_index.XSP  = 0x800000 -- characterごと
+stack_table_index.JWP  = 0 -- 0のみ
+stack_table_index.MJT  = 0x100 -- 0--255
+stack_table_index.MJS  = 0x200 -- 0--255
+stack_table_index.MJSS = 0x300 -- 0--255
+stack_table_index.KSJ  = 0x400 -- 0--9
+
+local userid_table = {}
+luatexja.userid_table = userid_table
+userid_table.IHB  = luatexbase.newuserwhatsitid('inhibitglue',  'luatexja') -- \inhibitglue
+userid_table.STCK = luatexbase.newuserwhatsitid('stack_marker', 'luatexja') -- スタック管理
+userid_table.OTF  = luatexbase.newuserwhatsitid('char_by_cid',  'luatexja') -- luatexja-otf
+userid_table.BPAR = luatexbase.newuserwhatsitid('begin_par',    'luatexja') -- 「段落始め」
+
+--- 定義終わり
+
 local load_module = luatexja.load_module
 load_module('base');      local ltjb = luatexja.base
 load_module('rmlgbm');    local ltjr = luatexja.rmlgbm -- must be 1st
@@ -47,28 +69,6 @@ load_module('setwidth');  local ltjw = luatexja.setwidth
 load_module('math');      local ltjm = luatexja.math
 load_module('tangle');    local ltjb = luatexja.base
 
-
-local node_type = node.type
-local node_new = node.new
-local node_prev = node.prev
-local node_next = node.next
-local has_attr = node.has_attribute
-local node_insert_before = node.insert_before
-local node_insert_after = node.insert_after
-local node_hpack = node.hpack
-
-local id_penalty = node.id('penalty')
-local id_glyph = node.id('glyph')
-local id_glue_spec = node.id('glue_spec')
-local id_glue = node.id('glue')
-local id_kern = node.id('kern')
-local id_hlist = node.id('hlist')
-local id_vlist = node.id('vlist')
-local id_rule = node.id('rule')
-local id_math = node.id('math')
-local id_whatsit = node.id('whatsit')
-local sid_user = node.subtype('user_defined')
-
 local attr_jchar_class = luatexbase.attributes['ltj@charclass']
 local attr_curjfnt = luatexbase.attributes['ltj@curjfnt']
 local attr_yablshift = luatexbase.attributes['ltj@yablshift']
@@ -76,20 +76,15 @@ local attr_icflag = luatexbase.attributes['ltj@icflag']
 local attr_uniqid = luatexbase.attributes['ltj@uniqid']
 local cat_lp = luatexbase.catcodetables['latex-package']
 
-local ITALIC = 1
-local PACKED = 2
-local KINSOKU = 3
-local FROM_JFM = 4
-local LINE_END = 5
-local KANJI_SKIP = 6
-local XKANJI_SKIP = 7
-local PROCESSED = 8
-local IC_PROCESSED = 9
-local BOXBDD = 15
+
+---- table: charprop_stack_table [stack_level].{pre|post|xsp}[chr_code]
 
 
 -- Three aux. functions, bollowed from tex.web
+
 local unity=65536
+local floor = math.floor
+
 local function print_scaled(s)
    local out=''
    local delta=10
@@ -131,10 +126,6 @@ local function print_spec(p)
 return out
 end
 
-function math.two_add(a,b) return a+b end
-function math.two_average(a,b) return (a+b)/2 end
-
----- table: charprop_stack_table [stack_level].{pre|post|xsp}[chr_code]
 
 ------------------------------------------------------------------------
 -- CODE FOR GETTING/SETTING PARAMETERS 
@@ -142,16 +133,17 @@ function math.two_average(a,b) return (a+b)/2 end
 
 -- EXT: print parameters that don't need arguments
 function luatexja.ext_get_parameter_unary(k)
+   local t = tex.getcount('ltj@@stack')
    if k == 'yalbaselineshift' then
       tex.write(print_scaled(tex.getattribute('ltj@yablshift'))..'pt')
    elseif k == 'yjabaselineshift' then
       tex.write(print_scaled(tex.getattribute('ltj@ykblshift'))..'pt')
    elseif k == 'kanjiskip' then
-      tex.write(print_spec(ltjs.get_skip_table('kanjiskip', tex.getcount('ltj@@stack'))))
+      tex.write(print_spec(ltjs.get_skip_table('kanjiskip', t)))
    elseif k == 'xkanjiskip' then
-      tex.write(print_spec(ltjs.get_skip_table('xkanjiskip', tex.getcount('ltj@@stack'))))
+      tex.write(print_spec(ltjs.get_skip_table('xkanjiskip', t)))
    elseif k == 'jcharwidowpenalty' then
-      tex.write(ltjs.get_penalty_table('jwp', 0, 0, tex.getcount('ltj@@stack')))
+      tex.write(ltjs.get_penalty_table(stack_table_index.JWP, 0, t))
    elseif k == 'autospacing' then
       tex.write(tex.getattribute('ltj@autospc'))
    elseif k == 'autoxspacing' then
@@ -163,6 +155,12 @@ function luatexja.ext_get_parameter_unary(k)
 	 tex.write('small')
       elseif luatexja.jfmglue.diffmet_rule == math.two_average then
 	 tex.write('average')
+      elseif luatexja.jfmglue.diffmet_rule == math.two_paverage then
+	 tex.write('paverage')
+      elseif luatexja.jfmglue.diffmet_rule == math.two_pleft then
+	 tex.write('pleft')
+      elseif luatexja.jfmglue.diffmet_rule == math.two_pright then
+	 tex.write('pright')
       elseif luatexja.jfmglue.diffmet_rule == math.two_add then
 	 tex.write('both')
       else -- This can't happen.
@@ -174,6 +172,7 @@ end
 
 -- EXT: print parameters that need arguments
 function luatexja.ext_get_parameter_binary(k,c)
+   local t = tex.getcount('ltj@@stack')
    if type(c)~='number' then
       ltjb.package_error('luatexja',
 			 'invalid the second argument (' .. tostring(c) .. ')',
@@ -181,14 +180,16 @@ function luatexja.ext_get_parameter_binary(k,c)
       c=0
    end
    if k == 'jacharrange' then
-      if c<0 or c>216 then 
+      if c>=31*ltjc.ATTR_RANGE then 
 	 ltjb.package_error('luatexja',
 			    'invalid character range number (' .. c .. ')',
-			    'A character range number should be in the range 0..216,\n'..
+			    'A character range number should be in the range 0..'
+                               .. 31*ltjc.ATTR_RANGE-1 .. ",\n"..
 			     'So I changed this one to zero.')
 	 c=0
       end
-      tex.write(ltjc.get_range_setting(c))
+      -- 負の値は <U+0080 の文字の文字範囲，として出てくる．この時はいつも欧文文字なので 1 を返す
+      tex.write( (c<0) and -1 or ltjc.get_range_setting(c))
    else
       if c<0 or c>0x10FFFF then
 	 ltjb.package_error('luatexja',
@@ -199,22 +200,22 @@ function luatexja.ext_get_parameter_binary(k,c)
 	 c=0
       end
       if k == 'prebreakpenalty' then
-	 tex.write(ltjs.get_penalty_table('pre', c, 0, tex.getcount('ltj@@stack')))
+	 tex.write(ltjs.get_penalty_table(stack_table_index.PRE + c, 0, t))
       elseif k == 'postbreakpenalty' then
-	 tex.write(ltjs.get_penalty_table('post', c, 0, tex.getcount('ltj@@stack')))
+	 tex.write(ltjs.get_penalty_table(stack_table_index.POST+ c, 0, t))
       elseif k == 'kcatcode' then
-	 tex.write(ltjs.get_penalty_table('kcat', c, 0, tex.getcount('ltj@@stack')))
+	 tex.write(ltjs.get_penalty_table(stack_table_index.KCAT+ c, 0, t))
       elseif k == 'chartorange' then 
 	 tex.write(ltjc.char_to_range(c))
       elseif k == 'jaxspmode' or k == 'alxspmode' then
-	 tex.write(ltjs.get_penalty_table('xsp', c, 3, tex.getcount('ltj@@stack')))
+	 tex.write(ltjs.get_penalty_table(stack_table_index.XSP + c, 3, t))
       end
    end
 end
 
 -- EXT: print \global if necessary
 function luatexja.ext_print_global()
-  if isglobal=='global' then tex.sprint(cat_lp, '\\global') end
+   if luatexja.isglobal=='global' then tex.sprint(cat_lp, '\\global') end
 end
 
 -- main process
@@ -233,25 +234,64 @@ luatexbase.add_to_callback('pre_linebreak_filter',
       return main_process(head, true, tex.textdir)
    end,'ltj.pre_linebreak_filter',
    luatexbase.priority_in_callback('pre_linebreak_filter',
-				   'luaotfload.pre_linebreak_filter') + 1)
+				   'luaotfload.node_processor') + 1)
 luatexbase.add_to_callback('hpack_filter', 
   function (head,groupcode,size,packtype, dir)
      return main_process(head, false, dir)
   end,'ltj.hpack_filter',
    luatexbase.priority_in_callback('hpack_filter',
-				   'luaotfload.hpack_filter') + 1)
+				   'luaotfload.node_processor') + 1)
+
+-- define_font
+
+local otfl_fdr = fonts.definers.read
+function luatexja.font_callback(name, size, id)
+  return ltjf.font_callback(
+     name, size, id, 
+     function (name, size, id) return ltjr.font_callback(name, size, id, otfl_fdr) end
+  )
+end
+--luatexbase.remove_from_callback('define_font',"luaotfload.define_font")
+luatexbase.add_to_callback('define_font',luatexja.font_callback,"luatexja.font_callback", 1)
+
+
+
+
 
 -- debug
+
+do
+
+local node_type = node.type
+local node_next = node.next
+local has_attr = node.has_attribute
+
+local id_penalty = node.id('penalty')
+local id_glyph = node.id('glyph')
+local id_glue_spec = node.id('glue_spec')
+local id_glue = node.id('glue')
+local id_kern = node.id('kern')
+local id_hlist = node.id('hlist')
+local id_vlist = node.id('vlist')
+local id_rule = node.id('rule')
+local id_math = node.id('math')
+local id_whatsit = node.id('whatsit')
+local sid_user = node.subtype('user_defined')
+
+local function get_attr_icflag(p)
+   return (has_attr(p, attr_icflag) or 0) % icflag_table.PROCESSED_BEGIN_FLAG
+end
+
 local debug_depth
 
 local function debug_show_node_X(p,print_fn)
    local k = debug_depth
    local s
    local pt=node_type(p.id)
-   local base = debug_depth .. string.format('%X', has_attr(p,attr_icflag) or 0)
-       .. ' ' .. tostring(p)
+   local base = debug_depth .. string.format('%X', get_attr_icflag(p))
+   .. ' ' .. pt .. ' ' .. tostring(p.subtype) .. ' '
    if pt == 'glyph' then
-      s = base .. ' ' .. utf.char(p.char) .. ' ' .. tostring(p.font)
+      s = base .. ' ' .. utf.char(p.char) .. ' '  .. tostring(p.font)
          .. ' (' .. print_scaled(p.height) .. '+' 
          .. print_scaled(p.depth) .. ')x' .. print_scaled(p.width)
       print_fn(s)
@@ -272,7 +312,7 @@ local function debug_show_node_X(p,print_fn)
             for i = 2,  p.glue_order do s = s .. 'l' end
          end
       end
-      if has_attr(p, attr_icflag, PACKED) then
+      if get_attr_icflag(p) == icflag_table.PACKED then
          s = s .. ' (packed)'
       end
       print_fn(s)
@@ -284,11 +324,12 @@ local function debug_show_node_X(p,print_fn)
       debug_depth=k
    elseif pt == 'glue' then
       s = base .. ' ' ..  print_spec(p.spec)
-      if has_attr(p, attr_icflag)==FROM_JFM then
-         s = s .. ' (from JFM)'
-      elseif has_attr(p, attr_icflag)==KANJI_SKIP then
+      if get_attr_icflag(p)>icflag_table.KINSOKU 
+         and get_attr_icflag(p)<icflag_table.KANJI_SKIP then
+         s = s .. ' (from JFM: priority ' .. get_attr_icflag(p)-icflag_table.FROM_JFM .. ')'
+      elseif get_attr_icflag(p)==icflag_table.KANJI_SKIP then
 	 s = s .. ' (kanjiskip)'
-      elseif has_attr(p, attr_icflag)==XKANJI_SKIP then
+      elseif get_attr_icflag(p)==icflag_table.XKANJI_SKIP then
 	 s = s .. ' (xkanjiskip)'
       end
       print_fn(s)
@@ -296,19 +337,18 @@ local function debug_show_node_X(p,print_fn)
       s = base .. ' ' .. print_scaled(p.kern) .. 'pt'
       if p.subtype==2 then
 	 s = s .. ' (for accent)'
-      elseif has_attr(p, attr_icflag)==IC_PROCESSED then
+      elseif get_attr_icflag(p)==icflag_table.IC_PROCESSED then
 	 s = s .. ' (italic correction)'
-         -- elseif has_attr(p, attr_icflag)==ITALIC then
+         -- elseif get_attr_icflag(p)==ITALIC then
          --    s = s .. ' (italic correction)'
-      elseif has_attr(p, attr_icflag)==FROM_JFM then
-	 s = s .. ' (from JFM)'
-      elseif has_attr(p, attr_icflag)==LINE_END then
-	 s = s .. " (from 'lineend' in JFM)"
+      elseif get_attr_icflag(p)>icflag_table.KINSOKU 
+         and get_attr_icflag(p)<icflag_table.KANJI_SKIP then
+	 s = s .. ' (from JFM: priority ' .. get_attr_icflag(p)-icflag_table.FROM_JFM .. ')'
       end
       print_fn(s)
    elseif pt == 'penalty' then
       s = base .. ' ' .. tostring(p.penalty)
-      if has_attr(p, attr_icflag)==KINSOKU then
+      if get_attr_icflag(p)==icflag_table.KINSOKU then
 	 s = s .. ' (for kinsoku)'
       end
       print_fn(s)
@@ -374,4 +414,6 @@ function luatexja.ext_show_node(head,depth,print_fn)
    else
       print_fn(debug_depth .. ' (null list)')
    end
-end
\ No newline at end of file
+end
+
+end
